@@ -140,10 +140,7 @@ struct insn *membase_reg_insn(enum insn_type insn_type, struct var_info *src_bas
 	return insn;
 }
 
-struct insn *memindex_insn(enum insn_type insn_type,
-			   struct var_info *src_base_reg,
-			   struct var_info *src_index_reg,
-			   unsigned char src_shift)
+struct insn *memindex_insn(enum insn_type insn_type, struct var_info *src_base_reg, struct var_info *src_index_reg, unsigned char src_shift)
 {
 	struct insn *insn = alloc_insn(insn_type);
 
@@ -153,6 +150,16 @@ struct insn *memindex_insn(enum insn_type insn_type,
 	return insn;
 }
 
+struct insn *reverse_memindex_insn(enum insn_type insn_type, struct var_info *dst_base_reg, struct var_info *dst_index_reg, unsigned char dst_shift)
+{
+	struct insn *insn = alloc_insn(insn_type);
+
+	if (insn)
+		init_memindex_operand(insn, &insn->dest, dst_base_reg, dst_index_reg, dst_shift);
+
+	return insn;
+}
+
 struct insn *memindex_reg_insn(enum insn_type insn_type,
 			       struct var_info *src_base_reg, struct var_info *src_index_reg,
 			       unsigned char src_shift, struct var_info *dest_reg)
@@ -401,6 +408,16 @@ struct insn *membase_insn(enum insn_type insn_type, struct var_info *src_base_re
 	return insn;
 }
 
+struct insn *reverse_membase_insn(enum insn_type insn_type, struct var_info *dst_base_reg, long dst_disp)
+{
+	struct insn *insn = alloc_insn(insn_type);
+
+	if (insn)
+		init_membase_operand(insn, &insn->dest, dst_base_reg, dst_disp);
+
+	return insn;
+}
+
 struct insn *imm_memlocal_insn(enum insn_type insn_type,
 			       unsigned long imm,
 			       struct stack_slot *dst_slot)
@@ -517,10 +534,10 @@ struct insn *spill_insn(struct var_info *var, struct stack_slot *slot)
 	assert(slot != NULL);
 
 	if (var->vm_type == J_FLOAT) {
-		insn_type = INSN_MOV_XMM_MEMLOCAL;
+		insn_type = INSN_MOVSS_XMM_MEMLOCAL;
 	} else if (var->vm_type == J_DOUBLE) {
 		if (cpu_has(X86_FEATURE_SSE2))
-			insn_type = INSN_MOV_64_XMM_MEMLOCAL;
+			insn_type = INSN_MOVSD_XMM_MEMLOCAL;
 		else
 			error("not implemented");
 	} else {
@@ -537,10 +554,10 @@ struct insn *reload_insn(struct stack_slot *slot, struct var_info *var)
 	assert(slot != NULL);
 
 	if (var->vm_type == J_FLOAT) {
-		insn_type = INSN_MOV_MEMLOCAL_XMM;
+		insn_type = INSN_MOVSS_MEMLOCAL_XMM;
 	} else if (var->vm_type == J_DOUBLE) {
 		if (cpu_has(X86_FEATURE_SSE2))
-			insn_type = INSN_MOV_64_MEMLOCAL_XMM;
+			insn_type = INSN_MOVSD_MEMLOCAL_XMM;
 		else
 			error("not implemented");
 	} else {
@@ -589,6 +606,9 @@ static unsigned long insn_flags[] = {
 	[INSN_ADC_IMM_REG]			= USE_DST | DEF_DST,
 	[INSN_ADC_MEMBASE_REG]			= USE_SRC | USE_DST | DEF_DST,
 	[INSN_ADC_REG_REG]			= USE_SRC | USE_DST | DEF_DST,
+	[INSN_ADDSD_MEMDISP_XMM]		= USE_DST | DEF_DST,
+	[INSN_ADDSD_XMM_XMM]			= USE_SRC | USE_DST | DEF_DST,
+	[INSN_ADDSS_XMM_XMM]			= USE_SRC | USE_DST | DEF_DST,
 	[INSN_ADD_IMM_REG]			= USE_DST | DEF_DST,
 	[INSN_ADD_MEMBASE_REG]			= USE_SRC | USE_DST | DEF_DST,
 	[INSN_ADD_REG_REG]			= USE_SRC | USE_DST | DEF_DST,
@@ -606,13 +626,10 @@ static unsigned long insn_flags[] = {
 	[INSN_CONV_GPR_TO_FPU]			= USE_SRC | DEF_DST,
 	[INSN_CONV_XMM64_TO_XMM]		= USE_SRC | DEF_DST,
 	[INSN_CONV_XMM_TO_XMM64]		= USE_SRC | DEF_DST,
+	[INSN_DIVSD_XMM_XMM]			= USE_SRC | USE_DST | DEF_DST,
+	[INSN_DIVSS_XMM_XMM]			= USE_SRC | USE_DST | DEF_DST,
 	[INSN_DIV_MEMBASE_REG]			= USE_SRC | USE_DST | DEF_DST | DEF_xAX | DEF_xDX,
 	[INSN_DIV_REG_REG]			= USE_SRC | USE_DST | DEF_DST | DEF_xAX | DEF_xDX,
-	[INSN_FADD_64_MEMDISP_REG]		= USE_DST | DEF_DST,
-	[INSN_FADD_64_REG_REG]			= USE_SRC | USE_DST | DEF_DST,
-	[INSN_FADD_REG_REG]			= USE_SRC | USE_DST | DEF_DST,
-	[INSN_FDIV_64_REG_REG]			= USE_SRC | USE_DST | DEF_DST,
-	[INSN_FDIV_REG_REG]			= USE_SRC | USE_DST | DEF_DST,
 	[INSN_FILD_64_MEMBASE]			= USE_SRC,
 	[INSN_FISTP_64_MEMBASE]			= USE_SRC | DEF_NONE,
 	[INSN_FLDCW_MEMBASE]			= USE_SRC | DEF_NONE,
@@ -620,51 +637,51 @@ static unsigned long insn_flags[] = {
 	[INSN_FLD_64_MEMLOCAL]			= USE_FP | DEF_NONE,
 	[INSN_FLD_MEMBASE]			= USE_SRC | DEF_NONE,
 	[INSN_FLD_MEMLOCAL]			= USE_FP | DEF_NONE,
-	[INSN_FMUL_64_MEMDISP_REG]		= USE_DST | DEF_DST,
-	[INSN_FMUL_64_REG_REG]			= USE_SRC | USE_DST | DEF_DST,
-	[INSN_FMUL_REG_REG]			= USE_SRC | USE_DST | DEF_DST,
 	[INSN_FNSTCW_MEMBASE]			= USE_SRC | DEF_NONE,
 	[INSN_FSTP_64_MEMBASE]			= USE_SRC | DEF_NONE,
 	[INSN_FSTP_64_MEMLOCAL]			= USE_FP | DEF_NONE,
 	[INSN_FSTP_MEMBASE]			= USE_SRC | DEF_NONE,
 	[INSN_FSTP_MEMLOCAL]			= USE_FP | DEF_NONE,
-	[INSN_FSUB_64_REG_REG]			= USE_SRC | USE_DST | DEF_DST,
-	[INSN_FSUB_REG_REG]			= USE_SRC | USE_DST | DEF_DST,
 	[INSN_JE_BRANCH]			= USE_NONE | DEF_NONE | TYPE_BRANCH,
 	[INSN_JGE_BRANCH]			= USE_NONE | DEF_NONE | TYPE_BRANCH,
 	[INSN_JG_BRANCH]			= USE_NONE | DEF_NONE | TYPE_BRANCH,
 	[INSN_JLE_BRANCH]			= USE_NONE | DEF_NONE | TYPE_BRANCH,
 	[INSN_JL_BRANCH]			= USE_NONE | DEF_NONE | TYPE_BRANCH,
 	[INSN_JMP_BRANCH]			= USE_NONE | DEF_NONE | TYPE_BRANCH,
-	[INSN_JMP_MEMBASE]			= USE_SRC | DEF_NONE | TYPE_BRANCH,
-	[INSN_JMP_MEMINDEX]			= USE_IDX_SRC | USE_SRC | DEF_NONE | TYPE_BRANCH,
+	[INSN_JMP_MEMBASE]			= USE_DST | DEF_NONE | TYPE_BRANCH,
+	[INSN_JMP_MEMINDEX]			= USE_IDX_DST | USE_DST | DEF_NONE | TYPE_BRANCH,
 	[INSN_JNE_BRANCH]			= USE_NONE | DEF_NONE | TYPE_BRANCH,
+	[INSN_MOVSD_MEMBASE_XMM]		= USE_SRC | DEF_DST,
+	[INSN_MOVSD_MEMDISP_XMM]		= USE_NONE | DEF_DST,
+	[INSN_MOVSD_MEMINDEX_XMM]		= USE_SRC | USE_IDX_SRC | DEF_DST,
+	[INSN_MOVSD_MEMLOCAL_XMM]		= USE_FP | DEF_DST,
+	[INSN_MOVSD_XMM_MEMBASE]		= USE_SRC | USE_DST,
+	[INSN_MOVSD_XMM_MEMDISP]		= USE_SRC | DEF_NONE,
+	[INSN_MOVSD_XMM_MEMINDEX]		= USE_SRC | USE_DST | USE_IDX_DST | DEF_NONE,
+	[INSN_MOVSD_XMM_MEMLOCAL]		= USE_SRC,
+	[INSN_MOVSD_XMM_XMM]			= USE_SRC | DEF_DST,
+	[INSN_MOVSS_MEMBASE_XMM]		= USE_SRC | DEF_DST,
+	[INSN_MOVSS_MEMDISP_XMM]		= USE_NONE | DEF_DST,
+	[INSN_MOVSS_MEMINDEX_XMM]		= USE_SRC | USE_IDX_SRC | DEF_DST,
+	[INSN_MOVSS_MEMLOCAL_XMM]		= USE_FP | DEF_DST,
+	[INSN_MOVSS_XMM_MEMBASE]		= USE_SRC | USE_DST,
+	[INSN_MOVSS_XMM_MEMDISP]		= USE_SRC | DEF_NONE,
+	[INSN_MOVSS_XMM_MEMINDEX]		= USE_SRC | USE_DST | USE_IDX_DST | DEF_NONE,
+	[INSN_MOVSS_XMM_MEMLOCAL]		= USE_SRC,
+	[INSN_MOVSS_XMM_XMM]			= USE_SRC | DEF_DST,
 	[INSN_MOVSX_16_MEMBASE_REG]		= USE_SRC | DEF_DST,
 	[INSN_MOVSX_16_REG_REG]			= USE_SRC | DEF_DST,
 	[INSN_MOVSX_8_MEMBASE_REG]		= USE_SRC | DEF_DST,
 	[INSN_MOVSX_8_REG_REG]			= USE_SRC | DEF_DST,
 	[INSN_MOVZX_16_REG_REG]			= USE_SRC | DEF_DST,
-	[INSN_MOV_64_MEMBASE_XMM]		= USE_SRC | DEF_DST,
-	[INSN_MOV_64_MEMDISP_XMM]		= USE_NONE | DEF_DST,
-	[INSN_MOV_64_MEMINDEX_XMM]		= USE_SRC | USE_IDX_SRC | DEF_DST,
-	[INSN_MOV_64_MEMLOCAL_XMM]		= USE_FP | DEF_DST,
-	[INSN_MOV_64_XMM_MEMBASE]		= USE_SRC | USE_DST,
-	[INSN_MOV_64_XMM_MEMDISP]		= USE_SRC | DEF_NONE,
-	[INSN_MOV_64_XMM_MEMINDEX]		= USE_SRC | USE_DST | USE_IDX_DST | DEF_NONE,
-	[INSN_MOV_64_XMM_MEMLOCAL]		= USE_SRC,
-	[INSN_MOV_64_XMM_XMM]			= USE_SRC | DEF_DST,
 	[INSN_MOV_IMM_MEMBASE]			= USE_DST,
 	[INSN_MOV_IMM_MEMLOCAL]			= USE_FP | DEF_NONE,
 	[INSN_MOV_IMM_REG]			= DEF_DST,
 	[INSN_MOV_IMM_THREAD_LOCAL_MEMBASE]	= USE_DST | DEF_NONE,
 	[INSN_MOV_MEMBASE_REG]			= USE_SRC | DEF_DST,
-	[INSN_MOV_MEMBASE_XMM]			= USE_SRC | DEF_DST,
 	[INSN_MOV_MEMDISP_REG]			= USE_NONE | DEF_DST,
-	[INSN_MOV_MEMDISP_XMM]			= USE_NONE | DEF_DST,
 	[INSN_MOV_MEMINDEX_REG]			= USE_SRC | USE_IDX_SRC | DEF_DST,
-	[INSN_MOV_MEMINDEX_XMM]			= USE_SRC | USE_IDX_SRC | DEF_DST,
 	[INSN_MOV_MEMLOCAL_REG]			= USE_FP | DEF_DST,
-	[INSN_MOV_MEMLOCAL_XMM]			= USE_FP | DEF_DST,
 	[INSN_MOV_REG_MEMBASE]			= USE_SRC | USE_DST,
 	[INSN_MOV_REG_MEMDISP]			= USE_SRC | DEF_NONE,
 	[INSN_MOV_REG_MEMINDEX]			= USE_SRC | USE_DST | USE_IDX_DST | DEF_NONE,
@@ -673,11 +690,9 @@ static unsigned long insn_flags[] = {
 	[INSN_MOV_REG_THREAD_LOCAL_MEMBASE]	= USE_SRC | USE_DST | DEF_NONE,
 	[INSN_MOV_REG_THREAD_LOCAL_MEMDISP]	= USE_SRC | DEF_NONE,
 	[INSN_MOV_THREAD_LOCAL_MEMDISP_REG]	= USE_NONE | DEF_DST,
-	[INSN_MOV_XMM_MEMBASE]			= USE_SRC | USE_DST,
-	[INSN_MOV_XMM_MEMDISP]			= USE_SRC | DEF_NONE,
-	[INSN_MOV_XMM_MEMINDEX]			= USE_SRC | USE_DST | USE_IDX_DST | DEF_NONE,
-	[INSN_MOV_XMM_MEMLOCAL]			= USE_SRC,
-	[INSN_MOV_XMM_XMM]			= USE_SRC | DEF_DST,
+	[INSN_MULSD_MEMDISP_XMM]		= USE_DST | DEF_DST,
+	[INSN_MULSD_XMM_XMM]			= USE_SRC | USE_DST | DEF_DST,
+	[INSN_MULSS_XMM_XMM]			= USE_SRC | USE_DST | DEF_DST,
 	[INSN_MUL_MEMBASE_EAX]			= USE_SRC | DEF_DST | DEF_xDX | DEF_xAX,
 	[INSN_MUL_REG_EAX]			= USE_SRC | USE_DST | DEF_DST | DEF_xDX | DEF_xAX,
 	[INSN_MUL_REG_REG]			= USE_SRC | USE_DST | DEF_DST,
@@ -698,15 +713,17 @@ static unsigned long insn_flags[] = {
 	[INSN_SBB_REG_REG]			= USE_SRC | USE_DST | DEF_DST,
 	[INSN_SHL_REG_REG]			= USE_SRC | USE_DST | DEF_DST,
 	[INSN_SHR_REG_REG]			= USE_SRC | USE_DST | DEF_DST,
+	[INSN_SUBSD_XMM_XMM]			= USE_SRC | USE_DST | DEF_DST,
+	[INSN_SUBSS_XMM_XMM]			= USE_SRC | USE_DST | DEF_DST,
 	[INSN_SUB_IMM_REG]			= USE_DST | DEF_DST,
 	[INSN_SUB_MEMBASE_REG]			= USE_SRC | USE_DST | DEF_DST,
 	[INSN_SUB_REG_REG]			= USE_SRC | USE_DST | DEF_DST,
 	[INSN_TEST_IMM_MEMDISP]			= USE_NONE | DEF_NONE,
 	[INSN_TEST_MEMBASE_REG]			= USE_SRC | USE_DST | DEF_NONE,
-	[INSN_XOR_64_XMM_REG_REG]		= USE_SRC | USE_DST | DEF_DST,
+	[INSN_XORPD_XMM_XMM]			= USE_SRC | USE_DST | DEF_DST,
 	[INSN_XOR_MEMBASE_REG]			= USE_SRC | USE_DST | DEF_DST,
 	[INSN_XOR_REG_REG]			= USE_SRC | USE_DST | DEF_DST,
-	[INSN_XOR_XMM_REG_REG]			= USE_SRC | USE_DST | DEF_DST,
+	[INSN_XORPS_XMM_XMM]			= USE_SRC | USE_DST | DEF_DST,
 };
 
 void insn_sanity_check(void)
