@@ -20,7 +20,7 @@
 
 #include "lib/buffer.h"
 
-struct basic_block *alloc_basic_block(struct compilation_unit *b_parent, unsigned long start, unsigned long end)
+struct basic_block *do_alloc_basic_block(struct compilation_unit *b_parent, unsigned long start, unsigned long end)
 {
 	struct basic_block *bb;
 
@@ -46,6 +46,17 @@ struct basic_block *alloc_basic_block(struct compilation_unit *b_parent, unsigne
 	return bb;
 }
 
+struct basic_block *alloc_basic_block(struct compilation_unit *b_parent, unsigned long start, unsigned long end)
+{
+	struct basic_block *ret;
+
+	ret	= do_alloc_basic_block(b_parent, start, end);
+	if (ret) {
+		b_parent->nr_bb++;
+	}
+	return ret;
+}
+
 struct basic_block *get_basic_block(struct compilation_unit *cu, unsigned long start, unsigned long end)
 {
 	struct basic_block *bb;
@@ -80,6 +91,7 @@ void shrink_basic_block(struct basic_block *bb)
 	free(bb->successors);
 	free(bb->predecessors);
 	free(bb->mimic_stack_expr);
+	free(bb->dom_frontier);
 	free(bb->use_set);
 	free(bb->def_set);
 	free(bb->live_in_set);
@@ -120,6 +132,8 @@ struct basic_block *bb_split(struct basic_block *orig_bb, unsigned long offset)
 
 	orig_bb->end = offset;
 
+	new_bb->entry_mimic_stack_size = -1;
+
 	new_bb->successors = orig_bb->successors;
 	orig_bb->successors = NULL;
 
