@@ -164,7 +164,7 @@ static void __spill_interval_intersecting(struct live_interval *current,
 {
 	unsigned long start;
 
-	if (it->reg != reg)
+	if (it->reg != (int8_t) reg)
 		return;
 
 	if (!intervals_intersect(it, current))
@@ -438,6 +438,7 @@ int allocate_registers(struct compilation_unit *cu)
 		}
 	}
 
+	pqueue_free(unhandled);
 	cu->is_reg_alloc_done = true;
 
 	return 0;
