@@ -31,6 +31,7 @@
 #include "vm/vm.h"
 
 #include "arch/stack-frame.h"
+#include "arch/registers.h"
 
 #include <ctype.h>
 #include <stdlib.h>
@@ -46,6 +47,7 @@ regex_t method_trace_regex;
 bool opt_trace_gate;
 regex_t method_trace_gate_regex;
 
+bool opt_trace_ssa;
 bool opt_trace_cfg;
 bool opt_trace_tree_ir;
 bool opt_trace_lir;
@@ -146,6 +148,49 @@ void trace_method(struct compilation_unit *cu)
 	trace_printf("\n\n");
 }
 
+static void print_bitset_frontier(struct bitset *bitset, struct compilation_unit *cu)
+{
+	bool exists = false;
+
+	for (unsigned int i = 0; i < bitset->nr_bits; ++i) {
+		if (test_bit(bitset->bits, i)){
+			if (!exists){
+				trace_printf("%d", i);
+				exists = true;
+			}else{
+				trace_printf(", %d", i);
+			}
+		}
+	}
+
+	if (!exists)
+		trace_printf("none    ");
+}
+
+void trace_ssa(struct compilation_unit *cu)
+{
+	struct basic_block *bb;
+
+	if (!cu_matches_regex(cu))
+		return;
+
+	trace_printf("SSA\n");
+	trace_printf("  #:\t\tDFN\tDominance frontier sets\n");
+
+	for_each_basic_block(bb, &cu->bb_list){
+		trace_printf("  %p\t%lu", bb, bb->dfn);
+		if (bb->is_eh)
+			trace_printf(" (eh)");
+
+		trace_printf("\t");
+		print_bitset_frontier(bb->dom_frontier, cu);
+		trace_printf("\n");
+	}
+
+	trace_printf("\n");
+
+}
+
 void trace_cfg(struct compilation_unit *cu)
 {
 	struct basic_block *bb;
@@ -509,11 +554,20 @@ static void print_array(struct vm_object *obj)
 	trace_printf("}");
 }
 
-static void print_arg(enum vm_type arg_type, const unsigned long *args,
-		      int *arg_index)
+static void print_arg(struct vm_method *vmm, enum vm_type arg_type,
+		      const unsigned long *args, int *arg_index)
 {
 	const int MAX_STRING_LENGTH = 50;
 
+	/*
+	 * We are passing pointer to struct vm_class as vmdata in
+	 * java.lang.Class(Object vmdata, ProtectionDomain pd)
+	 */
+	if (vmm == vm_java_lang_Class_init && *arg_index == 1) {
+		trace_printf("%p  (struct vm_class *)\n", args[(*arg_index)++]);
+		return;
+	}
+
 	if (arg_type == J_LONG || arg_type == J_DOUBLE) {
 		union {
 			unsigned long long ullvalue;
@@ -604,7 +658,7 @@ static void trace_invoke_args(struct vm_method *vmm,
 
 	if (!vm_method_is_static(vmm)) {
 		trace_printf("\tthis\t: ");
-		print_arg(J_REFERENCE, frame->args, &arg_index);
+		print_arg(vmm, J_REFERENCE, frame->args, &arg_index);
 	}
 
 	if (list_is_empty(&vmm->args)) {
@@ -617,7 +671,7 @@ static void trace_invoke_args(struct vm_method *vmm,
 	list_for_each_entry(arg, &vmm->args, list_node) {
 		trace_printf("\t   %-12s: ",
 			     get_vm_type_name(arg->type_info.vm_type));
-		print_arg(arg->type_info.vm_type, frame->args, &arg_index);
+		print_arg(vmm, arg->type_info.vm_type, frame->args, &arg_index);
 	}
 }
 
@@ -720,7 +774,7 @@ void trace_exception(struct compilation_unit *cu, struct jit_stack_frame *frame,
 	msg = field_get_object(exception, vm_java_lang_Throwable_detailMessage);
 
 	trace_printf("\tmessage\t: ");
-	print_arg(J_REFERENCE, (unsigned long *)  &msg, &dummy);
+	print_arg(vmm, J_REFERENCE, (unsigned long *)  &msg, &dummy);
 
 	trace_printf("\tfrom\t: %p: %s.%s%s\n", native_ptr, vmc->name, vmm->name,
 	       vmm->type);
@@ -811,6 +865,6 @@ void trace_return_value(struct vm_method *vmm, unsigned long long value)
 	}
 
 	trace_printf("%12s: ", get_vm_type_name(type));
-	print_arg(type,(unsigned long *)  &value, &dummy);
+	print_arg(vmm, type,(unsigned long *)  &value, &dummy);
 	trace_flush();
 }
