@@ -52,10 +52,13 @@ struct vm_class *vm_java_lang_Cloneable;
 struct vm_class *vm_java_lang_String;
 struct vm_class *vm_array_of_java_lang_String;
 struct vm_class *vm_java_lang_Throwable;
+struct vm_class *vm_java_util_HashMap;
 struct vm_class *vm_java_util_Properties;
 struct vm_class *vm_java_lang_VMThrowable;
 struct vm_class *vm_java_lang_StackTraceElement;
 struct vm_class *vm_array_of_java_lang_StackTraceElement;
+struct vm_class *vm_java_lang_Enum;
+struct vm_class *vm_array_of_java_lang_Enum;
 struct vm_class *vm_java_lang_Error;
 struct vm_class *vm_java_lang_InternalError;
 struct vm_class *vm_java_lang_OutOfMemoryError;
@@ -79,6 +82,7 @@ struct vm_class *vm_java_lang_InheritableThreadLocal;
 struct vm_class *vm_java_lang_VMThread;
 struct vm_class *vm_java_lang_IllegalMonitorStateException;
 struct vm_class *vm_java_lang_System;
+struct vm_class *vm_array_of_java_lang_annotation_Annotation;
 struct vm_class *vm_java_lang_reflect_Constructor;
 struct vm_class *vm_java_lang_reflect_Field;
 struct vm_class *vm_java_lang_reflect_Method;
@@ -88,6 +92,7 @@ struct vm_class *vm_java_lang_reflect_VMMethod;
 struct vm_class *vm_array_of_java_lang_reflect_Constructor;
 struct vm_class *vm_array_of_java_lang_reflect_Field;
 struct vm_class *vm_array_of_java_lang_reflect_Method;
+struct vm_class *vm_array_of_java_lang_Object;
 struct vm_class *vm_array_of_java_lang_Class;
 struct vm_class *vm_java_lang_IllegalArgumentException;
 struct vm_class *vm_java_lang_ClassLoader;
@@ -108,7 +113,9 @@ struct vm_class *vm_java_lang_ref_WeakReference;
 struct vm_class *vm_java_lang_ref_SoftReference;
 struct vm_class *vm_java_lang_ref_PhantomReference;
 struct vm_class *vm_java_nio_Buffer;
+struct vm_class *vm_java_nio_DirectByteBufferImpl_ReadWrite;
 struct vm_class *vm_gnu_classpath_PointerNN;
+struct vm_class *vm_sun_reflect_annotation_AnnotationInvocationHandler;
 struct vm_class *vm_boolean_class;
 struct vm_class *vm_char_class;
 struct vm_class *vm_float_class;
@@ -118,6 +125,14 @@ struct vm_class *vm_short_class;
 struct vm_class *vm_int_class;
 struct vm_class *vm_long_class;
 struct vm_class *vm_void_class;
+struct vm_class *vm_array_of_boolean;
+struct vm_class *vm_array_of_byte;
+struct vm_class *vm_array_of_char;
+struct vm_class *vm_array_of_double;
+struct vm_class *vm_array_of_float;
+struct vm_class *vm_array_of_int;
+struct vm_class *vm_array_of_long;
+struct vm_class *vm_array_of_short;
 
 static const struct preload_entry preload_entries[] = {
 	{ "java/lang/Object",		&vm_java_lang_Object },
@@ -125,7 +140,9 @@ static const struct preload_entry preload_entries[] = {
 	{ "java/lang/Cloneable",	&vm_java_lang_Cloneable },
 	{ "java/lang/String",		&vm_java_lang_String },
 	{ "[Ljava/lang/String;",	&vm_array_of_java_lang_String },
+	{ "[Ljava/lang/Enum;",		&vm_array_of_java_lang_Enum },
 	{ "java/lang/Throwable",	&vm_java_lang_Throwable },
+	{ "java/util/HashMap",		&vm_java_util_HashMap },
 	{ "java/util/Properties",	&vm_java_util_Properties },
 	{ "java/lang/StackTraceElement", &vm_java_lang_StackTraceElement },
 	{ "[Ljava/lang/StackTraceElement;", &vm_array_of_java_lang_StackTraceElement },
@@ -136,6 +153,7 @@ static const struct preload_entry preload_entries[] = {
 	{ "java/lang/ClassCastException", &vm_java_lang_ClassCastException },
 	{ "java/lang/ClassNotFoundException", &vm_java_lang_ClassNotFoundException },
 	{ "java/lang/Error",		&vm_java_lang_Error },
+	{ "java/lang/Enum",		&vm_java_lang_Enum },
 	{ "java/lang/InternalError",	 &vm_java_lang_InternalError },
 	{ "java/lang/OutOfMemoryError",	 &vm_java_lang_OutOfMemoryError },
 	{ "java/lang/ExceptionInInitializerError", &vm_java_lang_ExceptionInInitializerError },
@@ -153,12 +171,14 @@ static const struct preload_entry preload_entries[] = {
 	{ "java/lang/InheritableThreadLocal", &vm_java_lang_InheritableThreadLocal },
 	{ "java/lang/IllegalMonitorStateException", &vm_java_lang_IllegalMonitorStateException },
 	{ "java/lang/System",	&vm_java_lang_System },
+	{ "[Ljava/lang/annotation/Annotation;", &vm_array_of_java_lang_annotation_Annotation },
 	{ "java/lang/reflect/Field", &vm_java_lang_reflect_Field },
 	{ "java/lang/reflect/VMField", &vm_java_lang_reflect_VMField, PRELOAD_OPTIONAL }, /* Classpath 0.98 */
 	{ "java/lang/reflect/Constructor", &vm_java_lang_reflect_Constructor },
 	{ "java/lang/reflect/VMConstructor", &vm_java_lang_reflect_VMConstructor, PRELOAD_OPTIONAL }, /* Classpath 0.98 */
 	{ "java/lang/reflect/Method", &vm_java_lang_reflect_Method },
 	{ "java/lang/reflect/VMMethod", &vm_java_lang_reflect_VMMethod, PRELOAD_OPTIONAL }, /* Classpath 0.98 */
+	{ "[Ljava/lang/Object;",		&vm_array_of_java_lang_Object },
 	{ "[Ljava/lang/Class;",		&vm_array_of_java_lang_Class },
 	{ "[Ljava/lang/reflect/Constructor;", &vm_array_of_java_lang_reflect_Constructor },
 	{ "[Ljava/lang/reflect/Field;", &vm_array_of_java_lang_reflect_Field },
@@ -182,11 +202,13 @@ static const struct preload_entry preload_entries[] = {
 	{ "java/lang/ref/SoftReference", &vm_java_lang_ref_SoftReference },
 	{ "java/lang/ref/PhantomReference", &vm_java_lang_ref_PhantomReference },
 	{ "java/nio/Buffer", &vm_java_nio_Buffer },
+	{ "java/nio/DirectByteBufferImpl$ReadWrite", &vm_java_nio_DirectByteBufferImpl_ReadWrite },
 #ifdef CONFIG_32_BIT
 	{ "gnu/classpath/Pointer32", &vm_gnu_classpath_PointerNN },
 #else
 	{ "gnu/classpath/Pointer64", &vm_gnu_classpath_PointerNN },
 #endif
+	{ "sun/reflect/annotation/AnnotationInvocationHandler", &vm_sun_reflect_annotation_AnnotationInvocationHandler },
 };
 
 static const struct preload_entry primitive_preload_entries[] = {
@@ -201,6 +223,17 @@ static const struct preload_entry primitive_preload_entries[] = {
 	{"void", &vm_void_class},
 };
 
+static const struct preload_entry primitive_array_preload_entries[] = {
+	{"[Z", &vm_array_of_boolean},
+	{"[C", &vm_array_of_char},
+	{"[F", &vm_array_of_float},
+	{"[D", &vm_array_of_double},
+	{"[B", &vm_array_of_byte},
+	{"[S", &vm_array_of_short},
+	{"[I", &vm_array_of_int},
+	{"[J", &vm_array_of_long},
+};
+
 struct field_preload_entry {
 	struct vm_class **class;
 	const char *name;
@@ -325,6 +358,8 @@ struct method_preload_entry {
 	struct vm_method **method;
 };
 
+struct vm_method *vm_java_util_HashMap_init;
+struct vm_method *vm_java_util_HashMap_put;
 struct vm_method *vm_java_util_Properties_setProperty;
 struct vm_method *vm_java_lang_Throwable_initCause;
 struct vm_method *vm_java_lang_Throwable_getCause;
@@ -351,6 +386,7 @@ struct vm_method *vm_java_lang_Character_valueOf;
 struct vm_method *vm_java_lang_Class_init;
 struct vm_method *vm_java_lang_Double_init;
 struct vm_method *vm_java_lang_Double_valueOf;
+struct vm_method *vm_java_lang_Enum_valueOf;
 struct vm_method *vm_java_lang_Float_init;
 struct vm_method *vm_java_lang_Float_valueOf;
 struct vm_method *vm_java_lang_InheritableThreadLocal_newChildThread;
@@ -370,8 +406,22 @@ struct vm_method *vm_java_lang_Number_longValue;
 struct vm_method *vm_java_lang_Number_doubleValue;
 struct vm_method *vm_java_lang_ref_Reference_clear;
 struct vm_method *vm_java_lang_ref_Reference_enqueue;
+struct vm_method *vm_java_nio_DirectByteBufferImpl_ReadWrite_init;
+struct vm_method *vm_sun_reflect_annotation_AnnotationInvocationHandler_create;
 
 static const struct method_preload_entry method_preload_entries[] = {
+	{
+		&vm_java_util_HashMap,
+		"<init>",
+		"()V",
+		&vm_java_util_HashMap_init,
+	},
+	{
+		&vm_java_util_HashMap,
+		"put",
+		"(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;",
+		&vm_java_util_HashMap_put,
+	},
 	{
 		&vm_java_util_Properties,
 		"setProperty",
@@ -528,6 +578,12 @@ static const struct method_preload_entry method_preload_entries[] = {
 		"(D)Ljava/lang/Double;",
 		&vm_java_lang_Double_valueOf,
 	},
+	{
+		&vm_java_lang_Enum,
+		"valueOf",
+		"(Ljava/lang/Class;Ljava/lang/String;)Ljava/lang/Enum;",
+		&vm_java_lang_Enum_valueOf,
+	},
 	{
 		&vm_java_lang_Long,
 		"<init>",
@@ -642,6 +698,18 @@ static const struct method_preload_entry method_preload_entries[] = {
 		"()I",
 		&vm_java_lang_String_length,
 	},
+	{
+		&vm_java_nio_DirectByteBufferImpl_ReadWrite,
+		"<init>",
+		"(Ljava/lang/Object;Lgnu/classpath/Pointer;III)V",
+		&vm_java_nio_DirectByteBufferImpl_ReadWrite_init,
+	},
+	{
+		&vm_sun_reflect_annotation_AnnotationInvocationHandler,
+		"create",
+		"(Ljava/lang/Class;Ljava/util/Map;)Ljava/lang/annotation/Annotation;",
+		&vm_sun_reflect_annotation_AnnotationInvocationHandler_create,
+	},
 };
 
 /*
@@ -672,8 +740,6 @@ int vm_preload_add_class_fixup(struct vm_class *vmc)
 
 int preload_vm_classes(void)
 {
-	unsigned int array_size;
-
 	for (unsigned int i = 0; i < ARRAY_SIZE(preload_entries); ++i) {
 		const struct preload_entry *pe = &preload_entries[i];
 
@@ -688,8 +754,7 @@ int preload_vm_classes(void)
 		*pe->class = class;
 	}
 
-	array_size = ARRAY_SIZE(primitive_preload_entries);
-	for (unsigned int i = 0; i < array_size; ++i) {
+	for (unsigned int i = 0; i < ARRAY_SIZE(primitive_preload_entries); ++i) {
 		const struct preload_entry *pe = &primitive_preload_entries[i];
 
 		struct vm_class *class = classloader_load_primitive(pe->name);
@@ -703,6 +768,20 @@ int preload_vm_classes(void)
 		*pe->class = class;
 	}
 
+	for (unsigned int i = 0; i < ARRAY_SIZE(primitive_array_preload_entries); ++i) {
+		const struct preload_entry *pe = &primitive_array_preload_entries[i];
+
+		struct vm_class *class = classloader_load(NULL, pe->name);
+		if (!class) {
+			if (pe->optional == PRELOAD_OPTIONAL)
+				continue;
+			warn("preload of %s failed", pe->name);
+			return -EINVAL;
+		}
+
+		*pe->class = class;
+	}
+
 	for (unsigned int i = 0; i < ARRAY_SIZE(field_preload_entries); ++i) {
 		const struct field_preload_entry *pe = &field_preload_entries[i];
 
